"""
Gather everything to be created on worker process,
or not be able to be serialized,
or to be cached for all plugins
"""

from dask.distributed import WorkerPlugin, get_worker
import tflite_runtime.interpreter as tflite
import threading
import logging


# worker plugin to load the TensorFlow Lite model
class TensorFlowLiteModelPlugin(WorkerPlugin):
    def __init__(self, model_path):
        self.model_path = model_path
        self._log = logging.getLogger(__name__)

    def setup(self, worker):
        worker.plugin_instance = self
        worker.interpreters = {}

    def get_interpreter(self):
        worker = get_worker()
        thread_id = threading.get_ident()
        if thread_id not in worker.interpreters:
            interpreter = tflite.Interpreter(model_path=self.model_path)
            interpreter.allocate_tensors()
            worker.interpreters[thread_id] = interpreter
            self._log.info(
                "Interpreter created on worker %s for thread %s",
                worker.address,
                thread_id,
            )
        return worker.interpreters[thread_id]



"""
Plugin of Module A0 in LandCover PipeLine
"""

from functools import partial
from typing import Tuple, Dict, Iterable, Optional

import numpy as np
import xarray as xr
import dask.array as da
from odc.algo import keep_good_only
from odc.algo._masking import (
    _fuse_mean_np,
    _xr_fuse,
    mask_cleanup,
    to_float,
    _nodata_fuser,
)
from odc.stats._algebra import expr_eval, median_ds

from ._registry import StatsPluginInterface, register

NODATA = 255


class StatsVegCount(StatsPluginInterface):
    NAME = "ga_ls_lccs_fc_wo_a0"
    SHORT_NAME = NAME
    VERSION = "0.0.1"
    PRODUCT_FAMILY = "lccs"

    BAD_BITS_MASK = dict(cloud=(1 << 6), cloud_shadow=(1 << 5))

    def __init__(
        self,
        ue_threshold: Optional[int] = None,
        cloud_filters: Dict[str, Iterable[Tuple[str, int]]] = None,
        **kwargs,
    ):
        super().__init__(input_bands=["water", "pv", "bs", "npv", "ue"], **kwargs)

        self.ue_threshold = ue_threshold if ue_threshold is not None else 30
        self.cloud_filters = cloud_filters if cloud_filters is not None else {}

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurements = ["veg_frequency"]
        return _measurements

    def native_transform(self, xx):
        """
        Loads data in its native projection. It performs the following:

        1. Load all fc and WO bands
        3. Set all pixels that are not clear and dry to NODATA
        4. Calculate the clear wet pixels
        5. Drop the WOfS band
        """

        # clear and dry pixels not mask against bit 4: terrain high slope,
        # bit 3: terrain shadow, and
        # bit 2: low solar angle
        valid = (xx["water"] & ~((1 << 4) | (1 << 3) | (1 << 2))) == 0

        # clear and wet pixels not mask against bit 2: low solar angle
        wet = (xx["water"] & ~(1 << 2)) == 128

        # dilate both 'valid' and 'water'
        for key, val in self.BAD_BITS_MASK.items():
            if self.cloud_filters.get(key) is not None:
                raw_mask = (xx["water"] & val) > 0
                raw_mask = mask_cleanup(
                    raw_mask, mask_filters=self.cloud_filters.get(key)
                )
                valid &= ~raw_mask
                wet &= ~raw_mask

        xx = xx.drop_vars(["water"])

        # get valid wo pixels, both dry and wet
        data = expr_eval(
            "where(a|b, a, _nan)",
            dict(a=wet.data, b=valid.data),
            name="get_valid_pixels",
            dtype="float32",
            **dict(_nan=np.nan),
        )

        # Pick out the fc pixels that have an unmixing error of less than the threshold
        valid &= xx["ue"] < self.ue_threshold
        xx = xx.drop_vars(["ue"])
        xx = keep_good_only(xx, valid, nodata=NODATA)
        xx = to_float(xx, dtype="float32")

        xx["wet"] = xr.DataArray(data, dims=wet.dims, coords=wet.coords)

        return xx

    def fuser(self, xx):

        wet = xx["wet"]

        xx = _xr_fuse(xx.drop_vars(["wet"]), partial(_fuse_mean_np, nodata=np.nan), "")

        xx["wet"] = _nodata_fuser(wet, nodata=np.nan)

        return xx

    def _veg_or_not(self, xx: xr.Dataset):
        # either pv or npv > bs: 1
        # otherwise 0
        data = expr_eval(
            "where((a>b)|(c>b), 1, 0)",
            dict(a=xx["pv"].data, c=xx["npv"].data, b=xx["bs"].data),
            name="get_veg",
            dtype="uint8",
        )

        # mark nans
        data = expr_eval(
            "where(a!=a, nodata, b)",
            dict(a=xx["pv"].data, b=data),
            name="get_veg",
            dtype="uint8",
            **dict(nodata=int(NODATA)),
        )

        # mark water freq >= 0.5 as 0
        data = expr_eval(
            "where(a>0, 0, b)",
            dict(a=xx["wet"].data, b=data),
            name="get_veg",
            dtype="uint8",
        )

        return data

    def _max_consecutive_months(self, data, nodata):
        nan_mask = da.ones(data.shape[1:], chunks=data.chunks[1:], dtype="bool")
        tmp = da.zeros(data.shape[1:], chunks=data.chunks[1:], dtype="uint8")
        max_count = da.zeros(data.shape[1:], chunks=data.chunks[1:], dtype="uint8")

        for t in data:
            # +1 if not nodata
            tmp = expr_eval(
                "where(a==nodata, b, a+b)",
                dict(a=t, b=tmp),
                name="compute_consecutive_month",
                dtype="uint8",
                **dict(nodata=nodata),
            )

            # save the max
            max_count = expr_eval(
                "where(a>b, a, b)",
                dict(a=max_count, b=tmp),
                name="compute_consecutive_month",
                dtype="uint8",
            )

            # reset if not veg
            tmp = expr_eval(
                "where((a<=0), 0, b)",
                dict(a=t, b=tmp),
                name="compute_consecutive_month",
                dtype="uint8",
            )

            # mark nodata
            nan_mask = expr_eval(
                "where(a==nodata, b, False)",
                dict(a=t, b=nan_mask),
                name="mark_nodata",
                dtype="bool",
                **dict(nodata=nodata),
            )

        # mark nodata
        max_count = expr_eval(
            "where(a, nodata, b)",
            dict(a=nan_mask, b=max_count),
            name="mark_nodata",
            dtype="uint8",
            **dict(nodata=int(nodata)),
        )
        return max_count

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:

        xx = xx.groupby("time.month").map(median_ds, dim="spec")

        data = self._veg_or_not(xx)
        max_count = self._max_consecutive_months(data, NODATA)

        attrs = xx.attrs.copy()
        attrs["nodata"] = int(NODATA)
        data_vars = {
            "veg_frequency": xr.DataArray(
                max_count, dims=xx["wet"].dims[1:], attrs=attrs
            )
        }
        coords = dict((dim, xx.coords[dim]) for dim in xx["wet"].dims[1:])
        return xr.Dataset(data_vars=data_vars, coords=coords, attrs=xx.attrs)


register("veg_count", StatsVegCount)



# The content in this file will be moved to odc.algo after refactor.
# These are functionalities required to perform basic algebra in LandCover plugins
# Keep here for time being and an easier development/experiment

import dask.array as da
import xarray as xr
import numpy as np
import numexpr as ne
import functools
from dask.base import tokenize
from typing import Any, Dict, Optional
from odc.algo._dask import flatten_kv, unflatten_kv


def apply_numexpr_np(
    expr: str,
    data: Optional[Dict[str, Any]] = None,
    dtype=None,
    casting="safe",
    order="K",
    **params,
) -> np.ndarray:
    """
    Apply numexpr to numpy arrays
    """

    if data is None:
        data = params
    else:
        data.update(params)

    out = ne.evaluate(expr, local_dict=data, casting=casting, order=order)
    if dtype is None:
        return out
    else:
        return out.astype(dtype)


def expr_eval(expr, data, dtype="float32", name="expr_eval", **kwargs):
    tk = tokenize(apply_numexpr_np, *flatten_kv(data))
    op = functools.partial(
        apply_numexpr_np, expr, dtype=dtype, casting="unsafe", order="K", **kwargs
    )

    return da.map_blocks(
        lambda op, *data: op(unflatten_kv(data)),
        op,
        *flatten_kv(data),
        name=f"{name}_{tk}",
        dtype=dtype,
        meta=np.array((), dtype=dtype),
    )


def _median_by_ind(a):
    d = np.sort(a, axis=0).reshape(a.shape[0], -1)
    valid_count = d.shape[0] - np.isnan(d).sum(axis=0)
    indices = [valid_count // 2 - (1 - valid_count % 2), valid_count // 2]
    d = d.flatten(order="F")
    e = (
        d[indices[0] + np.arange(0, d.shape[0], a.shape[0])]
        + d[indices[1] + np.arange(0, d.shape[0], a.shape[0])]
    ) / 2
    return e.reshape(a.shape[1:])


def median_by_ind(xr_da, dim, dtype="float32", name="median_by_ind"):
    if xr_da.dims[0] != dim:
        raise ValueError(f"{dim} has to be on dimension 0")
    tk = tokenize(_median_by_ind, xr_da.data)
    res = da.map_blocks(
        _median_by_ind,
        xr_da.data,
        name=f"{name}_{tk}",
        dtype=dtype,
        meta=np.array((), dtype=dtype),
        drop_axis=0,
    )
    coords = dict((dim, xr_da.coords[dim]) for dim in xr_da.dims[1:])

    return xr.DataArray(
        res, dims=xr_da.dims[1:], coords=coords, attrs=xr_da.attrs.copy()
    )


def median_ds(xr_ds, dim, dtype="float32", name="median_ds"):
    res = {}
    for var, data in xr_ds.data_vars.items():
        res[var] = median_by_ind(data, dim, dtype, name)
    # pylint: disable=undefined-loop-variable
    coords = dict((dim, xr_ds.coords[dim]) for dim in data.dims[1:])
    return xr.Dataset(res, coords=coords, attrs=xr_ds.attrs.copy())



"""
Plugin of TF urban model in LandCover PipeLine
"""

from typing import Tuple, Dict, Sequence

import os
import numpy as np
import numexpr as ne
import xarray as xr
import dask.array as da
from dask.distributed import get_worker

from datacube.model import Dataset
from datacube.utils.geometry import GeoBox
from odc.algo._memsink import yxbt_sink
from odc.algo.io import load_with_native_transform

from odc.stats._algebra import expr_eval
from ._registry import StatsPluginInterface, register
from ._worker import TensorFlowLiteModelPlugin

NODATA = 255


# the padding and prediction function
def pad_and_predict(block, crop_size=(256, 256), nodata=NODATA):
    worker = get_worker()
    plugin_instance = worker.plugin_instance
    interpreter = plugin_instance.get_interpreter()

    if block.shape[0] < crop_size[0] or block.shape[1] < crop_size[1]:
        pad_width = [
            (0, max(0, crop_size[0] - block.shape[0])),
            (0, max(0, crop_size[1] - block.shape[1])),
        ] + [(0, 0)] * (block.ndim - 2)
        input_block = np.pad(block, pad_width, mode="edge")
    else:
        input_block = np.copy(block)

    # mark small holes as 0
    input_block = ne.evaluate(
        "where(a==a, a, 0)", local_dict={"a": input_block}
    ).astype("float32")
    # normalize pixels
    norm = np.linalg.norm(input_block, axis=-1, keepdims=True)
    input_block = input_block / np.maximum(norm, 1e-8)  # Avoid division by zero

    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()

    interpreter.set_tensor(
        input_details[0]["index"], input_block[np.newaxis, ...].astype(np.float32)
    )
    interpreter.invoke()

    output_data = interpreter.get_tensor(output_details[0]["index"])

    prediction = np.copy(output_data[0])
    prediction = np.squeeze(prediction, axis=-1)[0 : block.shape[0], 0 : block.shape[1]]
    prediction = ne.evaluate("where(a>0.5, 1, 0)", local_dict={"a": prediction})

    # mark small holes as nodata
    prediction = ne.evaluate(
        "where(a==a, b, nodata)",
        local_dict={"a": block[:, :, 0], "b": prediction, "nodata": nodata},
    ).astype("uint8")
    return prediction


class StatsUrbanClass(StatsPluginInterface):
    NAME = "ga_ls_urban"
    SHORT_NAME = NAME
    VERSION = "0.0.1"
    PRODUCT_FAMILY = "lccs"

    def __init__(
        self,
        output_classes: Dict,
        model_path: str,
        crop_size=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"{self.model_path} not found")
        self.dask_worker_plugin = TensorFlowLiteModelPlugin(model_path)
        self.output_classes = output_classes
        if crop_size is None:
            self.crop_size = (256, 256)
        else:
            self.crop_size = crop_size

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurements = ["urban_classes"]
        return _measurements

    def input_data(
        self, datasets: Sequence[Dataset], geobox: GeoBox, **kwargs
    ) -> xr.Dataset:
        # load data in the same time and location but different sensors
        data_vars = {}

        for ds in datasets:
            xx = load_with_native_transform(
                [ds],
                bands=self.input_bands,
                geobox=geobox,
                native_transform=self.native_transform,
                basis=self.basis,
                groupby=None,
                fuser=None,
                resampling=self.resampling,
                chunks=self.chunks,
                optional_bands=self.optional_bands,
                **kwargs,
            )
            input_array = yxbt_sink(
                xx, (self.crop_size[0], self.crop_size[0], -1, -1)
            ).squeeze("spec", drop=True)
            data_vars[ds.type.name] = input_array

        coords = dict((dim, input_array.coords[dim]) for dim in input_array.dims)
        return xr.Dataset(data_vars=data_vars, coords=coords)

    def urban_class(self, input_array):
        urban_mask = da.map_blocks(
            pad_and_predict,
            input_array,
            drop_axis=-1,
            dtype="uint8",
            crop_size=self.crop_size,
        )
        return urban_mask

    def impute_missing_values_from_group(self, xx):
        # Impute the missing values for each image
        images = []
        for img in xx.data_vars:
            image = xx[img].data
            nodata = xx[img].attrs.get("nodata", -999)
            for other_image in xx.data_vars:
                if other_image != img:
                    image = expr_eval(
                        "where((a<=nodata), b, a)",
                        {"a": image, "b": xx[other_image].data},
                        name="impute_missing_values",
                        **{
                            "nodata": nodata,
                        },
                    )

            # convert data type and mark nodata as nan
            image = expr_eval(
                "where((a<=nodata), _nan, a)",
                {
                    "a": image,
                },
                name="convert_dtype",
                dtype="float32",
                **{"nodata": nodata, "_nan": np.nan},
            )
            images += [image]
        return images

    def aggregate_results_from_group(self, urban_masks):
        # if there are >= 2 images
        # any is urban -> final class is urban
        # any is valid -> final class is valid
        # for each pixel
        m_size = len(urban_masks)
        if m_size > 1:
            urban_masks = da.stack(urban_masks).sum(axis=0)
        else:
            urban_masks = urban_masks[0]

        urban_masks = expr_eval(
            "where((a/nodata)>=_l, nodata, a%nodata)",
            {"a": urban_masks},
            name="mark_nodata",
            dtype="float32",
            **{"_l": m_size, "nodata": NODATA},
        )

        urban_masks = expr_eval(
            "where((a>0)&(a<nodata), _u, a)",
            {"a": urban_masks},
            name="output_classes_artificial",
            dtype="float32",
            **{
                "_u": self.output_classes["artificial"],
                "nodata": NODATA,
            },
        )

        urban_masks = expr_eval(
            "where(a<=0, _nu, a)",
            {"a": urban_masks},
            name="output_classes_natrual",
            dtype="uint8",
            **{
                "_nu": self.output_classes["natural"],
            },
        )

        return urban_masks.rechunk(-1, -1)

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        urban_masks = []
        images = self.impute_missing_values_from_group(xx)
        for image in images:
            urban_masks += [self.urban_class(image)]

        um = self.aggregate_results_from_group(urban_masks)

        attrs = xx.attrs.copy()
        attrs["nodata"] = int(NODATA)
        dims = list(xx.dims.keys())[:2]
        data_vars = {"urban_classes": xr.DataArray(um, dims=dims, attrs=attrs)}
        coords = {dim: xx.coords[dim] for dim in dims}
        return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)


register("urban_class", StatsUrbanClass)



"""
Plugin of Module A1 in LandCover PipeLine
"""

from typing import Tuple, Optional, Dict

import numpy as np
import xarray as xr
from odc.stats._algebra import expr_eval

from ._registry import StatsPluginInterface, register

NODATA = 255


class StatsDem(StatsPluginInterface):
    NAME = "ga_ls_lccs_(ni)dem"
    SHORT_NAME = NAME
    VERSION = "0.0.1"
    PRODUCT_FAMILY = "lccs"

    def __init__(
        self,
        measurements: Optional[Tuple[str, ...]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._measurements = (
            measurements if measurements is not None else self.input_bands
        )

    @property
    def measurements(self) -> Tuple[str, ...]:
        return self._measurements

    def native_transform(self, xx):
        # reproject cannot work with nodata being int for float
        # hence convert to nan

        for var in self.input_bands:
            nodata = float(xx[var].attrs["nodata"])
            data = expr_eval(
                "where(a>b, a, _nan)",
                {"a": xx[var].data},
                name="convert_nodata",
                dtype="float32",
                **{"_nan": np.nan, "b": nodata},
            )
            xx[var].data = data
            xx[var].attrs["nodata"] = np.nan

        return xx

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        if self.measurements != self.input_bands:
            xx = xx.rename(dict(zip(self.input_bands, self.measurements)))
        return xx


class StatsVegClassL1(StatsPluginInterface):
    NAME = "ga_ls_lccs_veg_class_a1"
    SHORT_NAME = NAME
    VERSION = "0.0.1"
    PRODUCT_FAMILY = "lccs"

    def __init__(
        self,
        output_classes: Dict,
        dem_threshold: Optional[int] = None,
        mudflat_threshold: Optional[int] = None,
        saltpan_threshold: Optional[int] = None,
        water_threshold: Optional[float] = None,
        veg_threshold: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.dem_threshold = dem_threshold if dem_threshold is not None else 6
        self.mudflat_threshold = (
            mudflat_threshold if mudflat_threshold is not None else 1000
        )
        self.saltpan_threshold = (
            saltpan_threshold if saltpan_threshold is not None else 1500
        )
        self.water_threshold = water_threshold if water_threshold is not None else 0.2
        self.veg_threshold = veg_threshold if veg_threshold is not None else 2
        self.output_classes = output_classes

    @property
    def measurements(self) -> Tuple[str, ...]:
        _measurements = ["classes_l3_l4"]
        return _measurements

    def native_transform(self, xx):
        return xx

    def fuser(self, xx):
        return xx

    def l3_class(self, xx: xr.Dataset):
        si5 = expr_eval(
            "where(a>nodata, a*b/c, _nan)",
            {"a": xx.nbart_blue.data, "b": xx.nbart_red.data, "c": xx.nbart_green.data},
            name="caculate_si5",
            dtype="float32",
            **{"_nan": np.nan, "nodata": xx.nbart_blue.attrs["nodata"]},
        )

        # water  (water_freq >= 0.2)

        l3_mask = expr_eval(
            "where((a>=wt), m, 0)",
            {"a": xx["frequency"].data},
            name="mark_water",
            dtype="uint8",
            **{
                "wt": self.water_threshold,
                "m": self.output_classes["water"],
            },
        )

        # surface: (si5 > 1000 & dem <=6 ) | (si5 > 1500) | (veg_freq < 2) & !water
        # rest:  aquatic/terretrial veg
        l3_mask = expr_eval(
            "where(((a>mt)&(b<=dt)|(a>st)|(d<vt))&(c<=0), m, c)",
            {
                "a": si5,
                "b": xx.dem_h.data,
                "d": xx.veg_frequency.data,
                "c": l3_mask,
            },
            name="mark_surface",
            dtype="uint8",
            **{
                "mt": self.mudflat_threshold,
                "dt": self.dem_threshold,
                "st": self.saltpan_threshold,
                "vt": self.veg_threshold,
                "m": self.output_classes["surface"],
            },
        )

        # if its mangrove or coast region
        for b in self.optional_bands:
            if b in xx.data_vars:
                if b == "elevation":
                    # intertidal: water | surface & elevation
                    # aquatic_veg: veg & elevation
                    data = expr_eval(
                        "where((a==a), 1, 0)",
                        {
                            "a": xx[b].data,
                        },
                        name="mark_intertidal",
                        dtype="bool",
                    )

                    l3_mask = expr_eval(
                        "where(a&(b>0), m, b)",
                        {"a": data, "b": l3_mask},
                        name="intertidal_water",
                        dtype="uint8",
                        **{"m": self.output_classes["intertidal"]},
                    )

                    l3_mask = expr_eval(
                        "where(a&(b<=0), m, b)",
                        {"a": data, "b": l3_mask},
                        name="intertidal_veg",
                        dtype="uint8",
                        **{"m": self.output_classes["aquatic_veg"]},
                    )
                elif b == "canopy_cover_class":
                    # aquatic_veg: (mangroves > 0) & (mangroves != nodata)
                    l3_mask = expr_eval(
                        "where((a>0)&(a<nodata), m, b)",
                        {"a": xx[b].data, "b": l3_mask},
                        name="mark_mangroves",
                        dtype="uint8",
                        **{"nodata": NODATA, "m": self.output_classes["aquatic_veg"]},
                    )

        # all unmarked values (0) is terretrial veg

        l3_mask = expr_eval(
            "where(a<=0, m, a)",
            {"a": l3_mask},
            name="mark_veg",
            dtype="uint8",
            **{"m": self.output_classes["terrestrial_veg"]},
        )

        # mark nodata if any source is nodata
        # issues:
        # - nodata information from non-indexed datasets missing

        l3_mask = expr_eval(
            "where((a!=a)|(b>=nodata), nodata, e)",
            {
                "a": si5,
                "b": xx.veg_frequency.data,
                "e": l3_mask,
            },
            name="mark_nodata",
            dtype="uint8",
            **{"nodata": NODATA},
        )

        return l3_mask

    def reduce(self, xx: xr.Dataset) -> xr.Dataset:
        l3_mask = self.l3_class(xx)

        attrs = xx.attrs.copy()
        attrs["nodata"] = int(NODATA)
        data_vars = {
            "classes_l3_l4": xr.DataArray(
                l3_mask[0], dims=xx["veg_frequency"].dims[1:], attrs=attrs
            )
        }
        coords = dict((dim, xx.coords[dim]) for dim in xx["veg_frequency"].dims[1:])
        return xr.Dataset(data_vars=data_vars, coords=coords, attrs=xx.attrs)


register("veg_class_l1", StatsVegClassL1)
register("dem_in_stats", StatsDem)
